"""
abc-classroom.config
====================

"""

import os
import sys
import pprint
from pathlib import Path
import os.path as op

from ruamel.yaml import YAML


def get_github_auth():
    """
    Check to see if there is an existing github authentication
    and load the authentication.

    Returns
    -------
    ruamel.yaml.comments.CommentedMap
        Yaml object that contains the token and id for a github session.
        If yaml doesn't exists, return an empty dictionary.
    """
    yaml = YAML()
    try:
        with open(op.expanduser("~/.abc-classroom.tokens.yml")) as f:
            config = yaml.load(f)
        return config["github"]

    except FileNotFoundError:
        return {}


def set_github_auth(auth_info):
    """
    Set the github authentication information. Put the token and id authentication
    information into a yaml file if it doesn't already exist.

    Parameters
    ----------
    auth_info : dictionary
        The token and id authentication information from github stored in a
        dictionary object.
    """
    yaml = YAML()
    config = {}
    if get_github_auth():
        with open(op.expanduser("~/.abc-classroom.tokens.yml")) as f:
            config = yaml.load(f)

    config["github"] = auth_info

    with open(op.expanduser("~/.abc-classroom.tokens.yml"), "w") as f:
        yaml.dump(config, f)


def get_config(configpath=None):
    yaml = YAML()
    try:
        if configpath is None:
            configpath = Path("config.yml")
        else:
            configpath = Path(configpath, "config.yml")
        with open(configpath) as f:
            config = yaml.load(f)
        return config
    except FileNotFoundError as err:
        configpath.resolve()
        print(
            "Oops! I can't seem to find a config.yml file at {}. "
            "Are you in the top-level directory for the course? If you don't have a course directory and config file "
            "setup yet, you can create one using abc-quickstart"
            ".\n".format(configpath)
        )
        sys.exit(1)


def print_config(config=None, configpath=None):
    """
    Print configuration. Can supply as dictionary parameter, or specify a path
    to look for config.yml. If neither option specified, looks for config.yml
    in current working directory. If both specified, prints dictionary, not
    from file.
    """
    configtoprint = {}
    if config is None:
        configtoprint = get_config(configpath)
    else:
        configtoprint = config
    print("Current configuration:\n")
    pprint.pprint(configtoprint)


def write_config(config, configpath=None):
    yaml = YAML()
    if configpath is None:
        configpath = Path("config.yml")
    else:
        configpath = Path(configpath, "config.yml")
    with open(configpath, "w") as f:
        yaml.dump(config, f)


# TODO: allow for nested gets, e.g. config[a][b]
def get_config_option(config, option, required=True):
    """
    Get an option (value of key) from provided config dictionary. If the key
    does not exist, exit with KeyError (required=True) or return None (required=False).
    """
    try:
        value = config[option]
        return value
    except KeyError as err:
        if required == True:
            print(
                "Did not find required option {} in config; exiting".format(
                    option
                )
            )
            sys.exit(1)
        else:
            return None


def set_config_option(
    config, option, value, append_value=False, configpath=None
):
    """
    Sets a config option. If option already exists and append_value is False,
    replaces existing value. If option exists and append_value is true, adds
    new value to list of existing values. Will not add a duplicate value.

    Writes the new config (overwriting the existing file) and returns new
    config dict.
    """

    existing_value = get_config_option(config, option, required=False)
    if append_value == True and existing_value is not None:
        if isinstance(existing_value, list):
            existing_value.append(value)
            value = existing_value
        else:
            value = [existing_value, value]
        # eliminate duplicates
        value = list(set(value))
    config[option] = value
    print("Writing new config at {}".format(configpath))
    write_config(config, configpath)
    return config



"""
abc-classroom.utils
===================

"""


import os
import datetime
import subprocess
import sys
import tempfile
import textwrap
from contextlib import contextmanager
from functools import lru_cache
from shutil import copystat, copy2

from IPython import get_ipython


class Error(OSError):
    pass


# a copy of shutil.copytree() that is ok with the target directory
# already existing
def copytree(
    src,
    dst,
    symlinks=False,
    ignore=None,
    copy_function=copy2,
    ignore_dangling_symlinks=False,
):
    """Recursively copy a directory tree.
    The destination directory must not already exist.
    If exception(s) occur, an Error is raised with a list of reasons.
    If the optional symlinks flag is true, symbolic links in the
    source tree result in symbolic links in the destination tree; if
    it is false, the contents of the files pointed to by symbolic
    links are copied. If the file pointed by the symlink doesn't
    exist, an exception will be added in the list of errors raised in
    an Error exception at the end of the copy process.
    You can set the optional ignore_dangling_symlinks flag to true if you
    want to silence this exception. Notice that this has no effect on
    platforms that don't support os.symlink.
    The optional ignore argument is a callable. If given, it
    is called with the `src` parameter, which is the directory
    being visited by copytree(), and `names` which is the list of
    `src` contents, as returned by os.listdir():
    ``callable(src, names) -> ignored_names``
    Since copytree() is called recursively, the callable will be
    called once for each directory that is copied. It returns a
    list of names relative to the `src` directory that should
    not be copied.
    The optional copy_function argument is a callable that will be used
    to copy each file. It will be called with the source path and the
    destination path as arguments. By default, copy2() is used, but any
    function that supports the same signature (like copy()) can be used.
    """
    names = os.listdir(src)
    if ignore is not None:
        ignored_names = ignore(src, names)
    else:
        ignored_names = set()

    os.makedirs(dst, exist_ok=True)
    errors = []
    for name in names:
        if name in ignored_names:
            continue
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if os.path.islink(srcname):
                linkto = os.readlink(srcname)
                if symlinks:
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(linkto, dstname)
                    copystat(srcname, dstname, follow_symlinks=not symlinks)
                else:
                    # ignore dangling symlink if the flag is on
                    if not os.path.exists(linkto) and ignore_dangling_symlinks:
                        continue
                    # otherwise let the copy occurs. copy2 will raise an error
                    if os.path.isdir(srcname):
                        copytree(
                            srcname, dstname, symlinks, ignore, copy_function
                        )
                    else:
                        copy_function(srcname, dstname)
            elif os.path.isdir(srcname):
                copytree(srcname, dstname, symlinks, ignore, copy_function)
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(srcname, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
    try:
        copystat(src, dst)
    except OSError as why:
        # Copying file access times may fail on Windows
        if getattr(why, "winerror", None) is None:
            errors.append((src, dst, str(why)))
    if errors:
        raise Error(errors)
    return dst


def input_editor(default_message=None):
    """Ask for user input via a text editor"""
    default_message = textwrap.dedent(default_message)

    with tempfile.NamedTemporaryFile(mode="r+") as tmpfile:
        if default_message is not None:
            tmpfile.write(default_message)
            tmpfile.flush()
        subprocess.check_call([get_editor(), tmpfile.name])
        tmpfile.seek(0)

        with open(tmpfile.name) as f:
            msg = f.read()
            return msg.strip()


def get_editor():
    return os.environ.get("VISUAL") or os.environ.get("EDITOR") or "vi"


@lru_cache(1)
def TOP():
    """Path to the top level of the repository we are in"""
    try:
        ret = _call_git("rev-parse", "--show-toplevel")
    except RuntimeError as e:
        print(" ".join(e.args))
        sys.exit(1)

    return ret.stdout.decode("utf-8").strip()


def P(*paths):
    """Construct absolute path inside the repository from `paths`"""
    path = os.path.join(*paths)
    return os.path.join(TOP(), path)


def get_abspath(testpath, coursepath):
    """
    Create an absoluate path of testpath inside coursepath if testpath is
    not already absolute.
    """
    if os.path.isabs(testpath):
        return testpath
    else:
        return os.path.join(coursepath, testpath)


def write_file(dir, filename, contents):
    """Write a new file called filename to directory dir.
    Each item in contents is a line in the file.
    """
    filepath = os.path.join(dir, filename)
    try:
        with open(filepath, "w") as f:
            for line in contents:
                f.write("{}\n".format(line))

    except OSError as err:
        print("Cannot open file: {0}".format(err))


def flush_inline_matplotlib_plots():
    """
    Flush matplotlib plots immediately, rather than asynchronously.
    Basically, the inline backend only shows the plot after the entire
    cell executes, which means we can't easily use a contextmanager to
    suppress displaying it. See https://github.com/jupyter-widgets/ipywidgets/issues/1181/
    and https://github.com/ipython/ipython/issues/10376 for more details. This
    function displays flushes any pending matplotlib plots if we are using
    the inline backend.

    Stolen from https://github.com/jupyter-widgets/ipywidgets/blob/4cc15e66d5e9e69dac8fc20d1eb1d7db825d7aa2/ipywidgets/widgets/interaction.py#L35
    """
    if "matplotlib" not in sys.modules:
        # matplotlib hasn't been imported, nothing to do.
        return

    try:
        import matplotlib as mpl
        from ipykernel.pylab.backend_inline import flush_figures
    except ImportError:
        return

    if mpl.get_backend() == "module://ipykernel.pylab.backend_inline":
        flush_figures()


def valid_date(s):
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d").date()
    except ValueError:
        msg = "Not a valid date: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


@contextmanager
def hide_outputs():
    """
    Context manager for hiding outputs from display() calls.

    IPython handles matplotlib outputs specially, so those are supressed too.
    """
    ipy = get_ipython()
    if ipy is None:
        # Not running inside ipython!
        yield
        return
    old_formatters = ipy.display_formatter.formatters
    ipy.display_formatter.formatters = {}
    try:
        yield
    finally:
        ipy.display_formatter.formatters = old_formatters


@contextmanager
def chdir(path):
    """Change working directory to `path` and restore old path on exit.
    `path` can be `None` in which case this is a no-op.
    """
    if path is None:
        yield
    else:
        old_dir = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(old_dir)



"""
abc-classroom.quickstart
========================

"""

import os
from pathlib import Path
from shutil import copy, rmtree
from . import config as cf
from abcclassroom import __file__


def path_to_example(dataset):
    """
    Construct a file path to an example dataset, assuming the dataset is
    contained in the 'example-data' directory of the abc-classroom package.
    Adapted from the PySAL package.

    Parameters
    ----------
    dataset: string
        Name of a dataset to access (e.g., "config.yml")

    Returns
    -------
    Path object
        A concrete Path object to the dataset
    """
    abcclassroom_path = Path(__file__).parent
    dataset_path = Path(abcclassroom_path, "example-data", dataset)
    if not dataset_path.exists():
        raise FileNotFoundError(dataset_path)
    return dataset_path


def create_dir_struct(course_name="abc_course", force=False, working_dir=None):
    """
    Create a directory structure that can be used to start an abc-classroom
    course. This includes a main directory, two sub directories for templates
    and cloned files, and a start to a configuration file.

    This is the tmplementation of the abc-quickstart script; it is called
    directly from main.
    """
    # Making sure the sample configuration file is where it's supposed to be.
    try:
        config_path = path_to_example("config.yml")
    except FileNotFoundError as err:
        print(
            """Sample config.yml configuration file cannot be located at {},
        please ensure abc-classroom has been installed
        correctly""".format(
                err
            )
        )
    # Assigning the custom folder name if applicable
    if " " in course_name:
        raise ValueError(
            "Spaces not allowed in custom course name: {}. Please use hyphens instead.".format(
                course_name
            )
        )
    if working_dir is None:
        working_dir = os.getcwd()
    main_dir = Path(working_dir, course_name)

    # Check if the main_dir doesn't exist already
    if main_dir.is_dir():
        if force:
            rmtree(main_dir)
        else:
            raise FileExistsError(
                """
                Ooops! It looks like the directory {} already exists on your
                computer. You might have already run quickstart in this directory.
                Consider using a different course name, deleting the existing
                directory, or running quikstart with the -f flag to force overwrite
                the existing directory.""".format(
                    course_name
                )
            )
    # make the main course directory and copy the config file there
    main_dir.mkdir()
    copy(config_path, main_dir)

    # now we can use config functions to read / write config
    config = cf.get_config(main_dir)
    config["course_directory"] = str(main_dir)
    cf.write_config(config, main_dir)
    clone_dir = cf.get_config_option(config, "clone_dir")
    template_dir = cf.get_config_option(config, "template_dir")

    # make the required subdirectories
    Path(main_dir, clone_dir).mkdir()
    Path(main_dir, template_dir).mkdir()

    print(
        """
        Created abc-classroom directory structure in '{}',
        including template and clone directories and a configuration file,
        'config.yml'. To proceed, please create / move your course roster
        and course_materials directory into '{}' and check the config file
        settings.""".format(
            main_dir, course_name
        )
    )



"""
abc-classroom.github
====================

"""

import os
import logging
import random
import string
import subprocess

import github3 as gh3

from .utils import input_editor


def _call_git(*args, directory=None):
    cmd = ["git"]
    cmd.extend(args)
    try:
        ret = subprocess.run(
            cmd,
            cwd=directory,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as e:
        err = e.stderr.decode("utf-8")
        if not err:
            err = e.stdout.decode("utf-8")
        raise RuntimeError(err) from e

    return ret


def remote_repo_exists(org, repository, token=None):
    """Check if the remote repository exists for the organization.
    """

    try:
        g = gh3.login(token=token)
        g.repository(org, repository)

    except Exception as e:
        return False

    return True


def check_student_repo_exists(org, course, student, token=None):
    """Check if the student has a repository for the course.

    It happens that students delete their repository or do not accept the
    invitation to the course. In either case they will not have a repository
    yet.
    """
    # temporarily change log level of github3.py as it prints weird messages
    # XXX could be done more nicely with a context manager maybe
    gh3_log = logging.getLogger("github3")
    old_level = gh3_log.level
    gh3_log.setLevel("ERROR")

    try:
        g = gh3.login(token=token)
        repository = "{}-{}".format(course, student)
        g.repository(org, repository)

    except Exception as e:
        raise e

    finally:
        gh3_log.setLevel(old_level)


def clone_repo(organization, repo, dest_dir):
    """Clone `repository` from `org` into a sub-directory in `directory`. Assumes you have ssh keys setup for github (rather than using GitHub API token).
    """
    url = "git@github.com:{}/{}.git".format(organization, repo)
    _call_git("-C", dest_dir, "clone", url)


def create_repo(org, repository, token):
    """Create a repository in the provided GitHub organization.
    """
    github_obj = gh3.login(token=token)
    organization = github_obj.organization(org)
    print(
        "Creating new repository {} at https://github.com/{}".format(
            repository, org
        )
    )
    try:
        organization.create_repository(repository)
    except gh3.exceptions.UnprocessableEntity as err:
        print(
            "Error: organization {} already has a repository named {}".format(
                org, repository
            )
        )


def add_remote(directory, organization, remote_repo, token):
    remote_url = "https://{}@github.com/{}/{}".format(
        token, organization, remote_repo
    )
    _call_git("remote", "add", "origin", remote_url, directory=directory)


def repo_changed(directory):
    """Determine if the Git repository in directory is dirty"""
    ret = _call_git("status", "--porcelain", directory=directory)
    return bool(ret.stdout)


def new_branch(directory, name=None):
    """Create a new git branch in directory"""
    if name is None:
        postfix = "".join(
            [random.choice(string.ascii_letters) for n in range(4)]
        )
        name = "new-material-{}".format(postfix)

    _call_git("checkout", "-b", name, directory=directory)

    return name


def get_commit_message():
    default_message = """
    # Please enter the commit message for your changes. Lines starting
    # with '#' will be ignored, and an empty message aborts the commit.
    # This message will be used as commit and Pull Request message."""
    message = input_editor(default_message)
    message = "\n".join(
        [
            line
            for line in message.split("\n")
            if not line.strip().startswith("#")
        ]
    )
    return message


def commit_all_changes(directory, msg=None):
    """Run git add, git commit on a given directory. Checks git status
    first and does nothing if no changes.
    """
    if msg is None:
        raise ValueError("Commit message can not be empty.")
    if repo_changed(directory):
        _call_git("add", "*", directory=directory)
        _call_git("commit", "-a", "-m", msg, directory=directory)
    else:
        print("No changes in repository {}; doing nothing".format(directory))


def init_and_commit(directory, custom_message=False):
    """Run git init, git add, git commit on given directory. Checks git status
    first and does nothing if no changes.
    """
    # local git things - initialize, add, commit
    # note that running git init on an existing repo is safe, so no need
    # to check anything first
    git_init(directory)
    if repo_changed(directory):
        message = "Initial commit"
        if custom_message:
            message = get_commit_message()
            if not message:
                print("Empty commit message, exiting.")
                sys.exit(1)
        commit_all_changes(directory, message)
    else:
        print("No changes to local repository.")


def push_to_github(directory, branch="master"):
    """Push `branch` back to GitHub"""
    try:
        _call_git(
            "push", "--set-upstream", "origin", branch, directory=directory
        )
    except RuntimeError as e:
        raise e


def pull_from_github(directory, branch="master"):
    """Pull `branch` of local repo in `directory` from GitHub"""
    try:
        _call_git("pull", "origin", branch, directory=directory)
    except RuntimeError as e:
        raise e


def git_init(directory):
    """Initialize git repository"""
    _call_git("init", directory=directory)


###################################################
# Methods below are from before the re-factoring.
# Retaining for reference, but with no guarantee
# about correct function.


def close_existing_pullrequests(
    org, repository, branch_base="new-material-", token=None
):
    """Close all oustanding course material update Pull Requests

    If there are any PRs open in a student's repository that originate from
    a branch starting with `branch_base` as name and created by the user
    we are logged in we close them.
    """
    g = gh3.login(token=token)
    me = g.me()
    repo = g.repository(org, repository)
    for pr in repo.pull_requests(state="open"):
        origin = pr.head.label
        origin_repo, origin_branch = origin.split(":")
        if origin_branch.startswith(branch_base) and pr.user == me:
            pr.create_comment(
                "Closed in favor of a new Pull Request to "
                "bring you up-to-date."
            )
            pr.close()


def create_pr(org, repository, branch, message, token):
    """Create a Pull Request with changes from branch"""
    msg_parts = message.split("\n\n")
    if len(msg_parts) == 1:
        title = msg = msg_parts[0]
    else:
        title = msg_parts[0]
        msg = "\n\n".join(msg_parts[1:])

    g = gh3.login(token=token)
    repo = g.repository(org, repository)
    repo.create_pull(title, "master", branch, msg)


def fetch_student(org, course, student, directory, token=None):
    """Fetch course repository for `student` from `org`

    The repository will be cloned into a sub-directory in `directory`.

    Returns the directory in which to find the students work.
    """
    # use ssh if there is no token
    if token is None:
        fetch_command = [
            "git",
            "clone",
            "git@github.com:{}/{}-{}.git".format(org, course, student),
        ]
    else:
        fetch_command = [
            "git",
            "clone",
            "https://{}@github.com/{}/{}-{}.git".format(
                token, org, course, student
            ),
        ]
    subprocess.run(
        fetch_command,
        cwd=directory,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    return os.path.join(directory, "{}-{}".format(course, student))



"""
abc-classroom.clone
======================

"""

import csv
from pathlib import Path
from shutil import copy

from . import config as cf
from . import github


def clone_or_update_repo(organization, repo, clone_dir, skip_existing):
    """
    Tries to clone the single repository 'repo' from the organization. If the
    local repository already exists, pulls instead of cloning (unless the
    skip flag is set, in which case it does nothing).
    """
    destination_dir = Path(clone_dir, repo)
    if destination_dir.is_dir():
        # if path exists, pull instead of clone (unless skip_existing)
        if skip_existing:
            print(
                "Local repo {} already exists; skipping".format(
                    destination_dir
                )
            )
            return
        github.pull_from_github(destination_dir)
    else:
        github.clone_repo(organization, repo, clone_dir)


def clone_student_repos(args):
    """Iterates through the student roster, clones each repo for this
    assignment into the directory specified in the config, and then copies the
    notebook files into the 'course_materials/submitted' directory, based on
    course_materials set in config.yml."""

    assignment = args.assignment
    skip_existing = args.skip_existing

    print("Loading configuration from config.yml")
    config = cf.get_config()
    roster_filename = cf.get_config_option(config, "roster", True)
    course_dir = cf.get_config_option(config, "course_directory", True)
    clone_dir = cf.get_config_option(config, "clone_dir", True)
    organization = cf.get_config_option(config, "organization", True)
    materials_dir = cf.get_config_option(config, "course_materials", False)

    if materials_dir is None:
        print(
            "No course_materials directory set in config.yml. Will clone "
            "repositories but will not copy any assignment files."
        )
    try:
        Path(course_dir, clone_dir).mkdir(exist_ok=True)
        missing = []
        with open(roster_filename, newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                student = row["github_username"]
                # expected columns: identifier,github_username,github_id,name
                repo = "{}-{}".format(assignment, student)
                try:
                    clone_or_update_repo(
                        organization, repo, clone_dir, skip_existing
                    )
                    if materials_dir is not None:
                        copy_assignment_files(config, student, assignment)
                except RuntimeError:
                    missing.append(repo)
        if len(missing) == 0:
            print("All successful; no missing repos")
        else:
            print("Could not clone or update the following repos: ")
            for r in missing:
                print(" {}".format(r))

    except FileNotFoundError as err:
        print("Cannot find roster file".format(roster_filename))
        print(err)


def copy_assignment_files(config, student, assignment):
    """Copies all notebook files from clone_dir to course_materials/submitted.
    Will overwrite any existing files with the same name."""
    course_dir = cf.get_config_option(config, "course_directory", True)
    materials_dir = cf.get_config_option(config, "course_materials", False)
    clone_dir = cf.get_config_option(config, "clone_dir", True)
    repo = "{}-{}".format(assignment, student)
    files = Path(course_dir, clone_dir, repo).glob("*.ipynb")
    destination = Path(
        course_dir, materials_dir, "submitted", student, assignment
    )
    destination.mkdir(parents=True, exist_ok=True)
    print(
        "Copying files from {} to {}".format(
            Path(clone_dir, repo), destination
        )
    )
    for f in files:
        print("copying {} to {}".format(f, destination))
        copy(f, destination)



import argparse
import datetime
import glob
import os
import shutil
import subprocess
import sys
import tempfile

import os.path as op
from argparse import ArgumentParser
from getpass import getpass

import github3 as gh3

from . import ok
from . import template
from . import feedback as fdback
from . import config as cf
from .distribute import find_notebooks, render_circleci_template
from .notebook import split_notebook
from .quickstart import create_dir_struct
from .clone import clone_student_repos
from . import github
from .utils import copytree, P, input_editor, write_file, valid_date


def quickstart():
    parser = ArgumentParser(description=create_dir_struct.__doc__)
    parser.add_argument(
        "course_name",
        nargs="?",
        default="course-directory",
        help="Name of course. Use dashes-rather-than spaces for your name.",
    )
    parser.add_argument(
        "-f",
        action="store_true",
        help="Option to override the existing folder structure made by this function previously.",
    )
    args = parser.parse_args()
    course_name = args.course_name
    create_dir_struct(course_name, args.f)


def init():
    """
    Setup GitHub credentials for later. Make sure that there is a valid GitHub authentication
    yaml file, and if there isn't one, create a valid file.
    """

    gh_auth = cf.get_github_auth()

    # check the token we have is still valid by attempting to login with
    # the token we have if this fails we need a new one
    if gh_auth.get("token") is not None:
        try:
            # We have to use the GitHub API to find out if our login
            # credentials actually work, this is why we call `me()`
            gh = gh3.login(token=gh_auth["token"])
            gh.me()
            print("GitHub token present and valid.")
            return

        except gh3.exceptions.AuthenticationFailed:
            # need to get a new token
            pass

    print("GitHub token is missing or expired. Populating")
    user = input("GitHub username: ")
    password = ""

    while not password:
        password = getpass("Password for {0}: ".format(user))

    note = "ABC-classroom workflow helper"
    note_url = "https://github.com/earthlab/abc-classroom"
    scopes = ["repo", "read:user"]

    def two_factor():
        code = ""
        while not code:
            # The user could accidentally press Enter before being ready,
            # let's protect them from doing that.
            code = input("Enter 2FA code: ")
        return code

    gh = gh3.github.GitHub()
    gh.login(username=user, password=password, two_factor_callback=two_factor)
    try:
        auth = gh.authorize(user, password, scopes, note, note_url)

        cf.set_github_auth({"token": auth.token, "id": auth.id})

    except gh3.exceptions.UnprocessableEntity:
        print(
            "Failed to create a access token for you. Please visit "
            "https://github.com/settings/tokens and delete any access "
            "token with the name 'ABC-classroom workflow helper' and run "
            "`abc-init` again."
        )
        sys.exit(1)


def clone():
    """
    Clone the student repositories for the assignment and (optionall) copies
    notebook files into the course_materials 'submitted' directory. Clones into
    the clone_dir directory, as specified in config.yml.

    Requires that you have filename of student roster
    defined in config.yml and that the roster file exists.

    By default, if a local directory with the name of the repo already exists,
    pulls from github to update. Use the --skip-existing flag if you don't want
    to update existing repos.
    """
    parser = argparse.ArgumentParser(description=clone.__doc__)
    parser.add_argument(
        "assignment",
        help="Name of assignment. Must match assignment name in course_materials directories",
    )
    parser.add_argument(
        "--skip-existing",
        action="store_true",
        help="Do not attempt to update repositories that have already been cloned.",
    )
    args = parser.parse_args()

    clone_student_repos(args)


def feedback():
    """
    Copies feedback reports to local student repositories and (optionally) pushes to github. Assumes files are in the directory course_materials/feedback/student/assignment. Copies all files in the source directory.
    """
    parser = argparse.ArgumentParser(description=feedback.__doc__)
    parser.add_argument(
        "assignment",
        help="Name of assignment. Must match name in course_materials feedback directory",
    )
    parser.add_argument(
        "--github",
        action="store_true",
        help="Also pushes files to student repositories on GitHub (default = False; only copies files to local repos)",
    )
    args = parser.parse_args()
    fdback.copy_feedback(args)


def new_template():
    """
    Create a new assignment template repository: creates local directory,
    copy / create required files, intialize as git repo, and (optionally) create remote repo
    on GitHub and push local repo to GitHub. Will open git editor to ask for
    commit message if custom message requested.
    """
    parser = argparse.ArgumentParser(description=new_template.__doc__)
    parser.add_argument(
        "assignment",
        help="Name of assignment. Must match name in course_materials/release directory",
    )
    parser.add_argument(
        "--custom-message",
        action="store_true",
        help="Use a custom commit message for git. Will open the default git text editor for entry (if not set, uses default message 'Initial commit').",
    )
    parser.add_argument(
        "--github",
        action="store_true",
        help="Also perform the GitHub operations (create remote repo on GitHub and push to remote (by default, only does local repository setup)",
    )
    parser.add_argument(
        "--mode",
        choices=["delete", "fail", "merge"],
        default="fail",
        help="Action if template directory already exists. Choices are: delete = delete contents before proceeding (except .git directory); merge = keep existing dir, overwrite existing files, add new files (Default = fail).",
    )
    args = parser.parse_args()

    template.new_update_template(args)


def update_template():
    """
    Updates an existing assignment template repository: update / add new and  changed files, then push local changes to GitHub. Will open git editor to ask for
    commit message.
    """
    parser = argparse.ArgumentParser(description=update_template.__doc__)
    parser.add_argument(
        "assignment",
        help="Name of assignment. Must match name in course_materials/release directory",
    )
    parser.add_argument(
        "--mode",
        choices=["delete", "merge"],
        default="merge",
        help="What to do with existing contents of template directory. Choices are: delete = remove contents before proceeding (leaving .git directory); merge = overwrite existing files add new files (Default = merge).",
    )
    args = parser.parse_args()
    # now set the additional args (so that it matches the keys in add_template and we can use the same implementation
    # methods)
    setattr(args, "github", True)
    setattr(args, "custom_message", True)
    template.new_update_template(args)


#########################################
# Everything below here - grade, author, and distribute methods - is from
# before the re-factor, when grading happened
# within abcclassroom. Keeping for reference, but with no guarantees that
# this code is functional.
########################################

# def grade():
#     """Grade student's work"""
#     parser = argparse.ArgumentParser(description="Grade student repository.")
#     parser.add_argument(
#         "--date",
#         default=datetime.datetime.today().date(),
#         type=valid_date,
#         help=("Assumed date when grading assignments " "(default: today)"),
#     )
#     parser.add_argument(
#         "--student",
#         default=None,
#         action="append",
#         help=(
#             "Student name to grade, use flag multiple times "
#             "to select several students "
#             "(default: all students)"
#         ),
#     )
#     parser.add_argument(
#         "--assignment",
#         default=None,
#         action="append",
#         help=(
#             "Assignment to grade, use flag multiple times "
#             "to select several assignments "
#             "(default: all assignments)"
#         ),
#     )
#     args = parser.parse_args()
#
#     now = args.date
#
#     config = get_config()
#     course = config["courseName"]
#
#     if args.student is None:
#         students = config["students"]
#     else:
#         students = args.student
#
#     if args.assignment is None:
#         assignments = config["assignments"]
#     else:
#         assignments = args.assignment
#
#     for student in students:
#         print("Fetching work for %s..." % student)
#         cwd = P("graded", student)
#
#         # always delete and recreate student's directories
#         if os.path.exists(cwd):
#             shutil.rmtree(cwd)
#         os.makedirs(cwd)
#
#         # Use HTTPS and tokens to avoid access problems
#         # `git clone https://<token>@github.com/owner/repo.git`
#         fetch_command = [
#             "git",
#             "clone",
#             "https://{}@github.com/{}/{}-{}.git".format(
#                 get_github_auth()["token"],
#                 config["organisation"],
#                 course,
#                 student,
#             ),
#             student,
#         ]
#         try:
#             subprocess.run(
#                 fetch_command,
#                 cwd=P("graded"),
#                 check=True,
#                 stdout=subprocess.PIPE,
#                 stderr=subprocess.PIPE,
#             )
#         except subprocess.CalledProcessError as err:
#             print("Fetching work with '%s' failed:" % " ".join(fetch_command))
#             print()
#             print(err.stderr.decode("utf-8"))
#             print()
#             print("Skipping this student.")
#             continue
#
#         print("Grading work...")
#         for assignment in assignments:
#             # only grade assignments that are due or have been explicitly
#             # requested by command-line flag
#             deadline_date = config["assignments"][assignment]["deadline"]
#             if deadline_date > now and args.assignment is None:
#                 print('Skipping assignment "{}".'.format(assignment))
#
#             # remove check files so we only use a clean copy from this repo
#             # instead of trusting students
#             for notebook in glob.glob(
#                 P("graded", student, "%s/*.ipynb" % assignment)
#             ):
#                 print("Grading {}".format(notebook))
#                 notebook = op.split(notebook)[-1]
#
#                 tests_path = P(
#                     "graded", student, assignment, op.splitext(notebook)[0]
#                 )
#                 if os.path.exists(tests_path):
#                     shutil.rmtree(tests_path)
#
#                 autograder_path = P(
#                     "autograder", assignment, op.splitext(notebook)[0]
#                 )
#                 copytree(autograder_path, tests_path)
#
#                 results = ok.grade_notebook(
#                     P("graded", student, assignment, notebook)
#                 )
#
#                 print("Points:")
#                 for res in results:
#                     print(res)
#
#             print(
#                 "🎉 Top marks for {} on assignment {}.".format(
#                     student, assignment
#                 )
#             )
#         print()
#
#
# def distribute():
#     """Create or update student repositories"""
#     parser = argparse.ArgumentParser(description="Distribute work to students")
#     parser.add_argument(
#         "--template",
#         action="store_true",
#         help="Create template repository only (default: False)",
#     )
#     args = parser.parse_args()
#
#     student_repo_template = P("student")
#
#     print("Using %s to create the student template." % student_repo_template)
#     print("Loading configuration from config.yml")
#
#     config = get_config()
#
#     if args.template:
#         print("Creating template repository.")
#         repo_name = "{}-{}".format(config["courseName"], "template")
#         with tempfile.TemporaryDirectory() as d:
#             copytree(P("student"), d)
#             gitu.git_init(d)
#             gitu.commit_all_changes(d, "Initial commit")
#             try:
#                 gitu.create_repo(
#                     config["organisation"],
#                     repo_name,
#                     d,
#                     get_github_auth()["token"],
#                 )
#             except gh3.exceptions.UnprocessableEntity as e:
#                 print(e.msg)
#                 print(
#                     "This is probably because the template repository "
#                     "already exists."
#                 )
#
#         print(
#             "Visit https://github.com/{}/{}".format(
#                 config["organisation"], repo_name
#             )
#         )
#
#     else:
#         default_message = """
#         # Please enter the commit message for your changes. Lines starting
#         # with '#' will be ignored, and an empty message aborts the commit.
#         # This message will be used as commit and Pull Request message."""
#         message = input_editor(default_message)
#         message = "\n".join(
#             [
#                 line
#                 for line in message.split("\n")
#                 if not line.strip().startswith("#")
#             ]
#         )
#
#         if not message:
#             print("Empty commit message, exiting.")
#             sys.exit(1)
#
#         for student in config["students"]:
#             print("Fetching work for %s..." % student)
#
#             try:
#                 gitu.check_student_repo_exists(
#                     config["organisation"],
#                     config["courseName"],
#                     student,
#                     token=get_github_auth()["token"],
#                 )
#             except gh3.exceptions.NotFoundError as e:
#                 print(
#                     "Student {} does not have a repository for this "
#                     "course, maybe they have not accepted the invitation "
#                     "yet? Skipping them for now.".format(student)
#                 )
#                 continue
#
#             with tempfile.TemporaryDirectory() as d:
#                 student_dir = gitu.fetch_student(
#                     config["organisation"],
#                     config["courseName"],
#                     student,
#                     directory=d,
#                     token=get_github_auth()["token"],
#                 )
#                 # Copy assignment related files to the template repository
#                 copytree(P("student"), student_dir)
#
#                 if gitu.repo_changed(student_dir):
#                     # only close outstanding PRs if we are about to make a
#                     # new PR. Otherwise we can skip this.
#                     repo = "{}-{}".format(config["courseName"], student)
#                     gitu.close_existing_pullrequests(
#                         config["organisation"],
#                         repo,
#                         token=get_github_auth()["token"],
#                     )
#
#                     branch = gitu.new_branch(student_dir)
#
#                     gitu.commit_all_changes(student_dir, message)
#                     gitu.push_to_github(student_dir, branch)
#                     gitu.create_pr(
#                         config["organisation"],
#                         repo,
#                         branch,
#                         message,
#                         get_github_auth()["token"],
#                     )
#
#                 else:
#                     print("Everything up to date.")
#
#
# def author():
#     """Create student repository and autograding tests"""
#     parser = argparse.ArgumentParser(description="Author student repository.")
#     parser.add_argument(
#         "--date",
#         default=datetime.datetime.today().date(),
#         type=valid_date,
#         help=("Assumed date when preparing assignments " "(default: today)"),
#     )
#     args = parser.parse_args()
#
#     config = get_config()
#     now = args.date
#
#     if os.path.exists(P("student")):
#         shutil.rmtree(P("student"))
#     os.makedirs(P("student"))
#
#     if os.path.exists(P("autograder")):
#         shutil.rmtree(P("autograder"))
#
#     for assignment in config["assignments"]:
#         release_date = config["assignments"][assignment]["release"]
#         if release_date > now:
#             continue
#
#         student_path = P("student", assignment)
#         master_path = P("master", assignment)
#
#         if not os.path.isdir(master_path):
#             print(
#                 "Error: There is no material in '{}' for "
#                 "assignment '{}'".format(master_path, assignment)
#             )
#             sys.exit(1)
#
#         # copy over everything, including master notebooks. They will be
#         # overwritten by split_notebook() below
#         shutil.copytree(master_path, student_path)
#
#         if os.path.exists(P("student", assignment, ".ipynb_checkpoints")):
#             shutil.rmtree(P("student", assignment, ".ipynb_checkpoints"))
#
#         for notebook in glob.glob(P("master/%s/*.ipynb" % assignment)):
#             split_notebook(notebook, student_path, P("autograder", assignment))
#
#     # Create additional files
#     for target, source in config["extra_files"].items():
#         shutil.copyfile(P(source), P("student", target))
#
#     # Create the grading token file which is used by the notebook bot
#     # to access the CircleCI build artefacts
#     grading_token = P("student", ".grading.token")
#     with open(grading_token, "w") as f:
#         f.write(config["tokens"]["circleci"])
#
#     # Create the required CircleCI configuration
#     # the template only needs the basename, not the .ipynb extension
#     notebook_paths = [f[:-6] for f in find_notebooks(P("student"))]
#     circleci = render_circleci_template(notebook_paths)
#
#     os.makedirs(P("student", ".circleci"))
#     circleci_yml = P("student", ".circleci", "config.yml")
#     with open(circleci_yml, "w") as f:
#         f.write(circleci)
#
#     print(
#         "Inspect `{}/` to check it looks as you "
#         "expect.".format(P("student"))
#     )



"""
abc-classroom.notebook
======================

"""

import ast
import os

import nbformat
import papermill as pm

from nbclean import NotebookCleaner

from .utils import chdir


try:
    from IPython.core.inputsplitter import IPythonInputSplitter
except ImportError:
    raise ImportError("IPython needs to be installed for notebook grading")


def normalize_kernel_name(notebook):
    nb = nbformat.read(notebook, as_version=4)

    kernelspec = nb.metadata.kernelspec
    if "[conda env:" in kernelspec.display_name:
        if kernelspec.language == "python":
            if nb.metadata.language_info.version.startswith("3."):
                kernelspec.name = "python3"
                kernelspec.display_name = "Python 3"
            else:
                kernelspec.name = "python2"
                kernelspec.display_name = "Python 2"

        nbformat.write(nb, notebook)


def split_notebook(notebook, student_path, autograder_path):
    """Split a master notebook into student and autograder notebooks"""
    print("Processing", notebook)

    _, nb_name = os.path.split(notebook)
    base_name, extension = os.path.splitext(nb_name)

    # create test files and notebook for the student
    nb = NotebookCleaner(notebook)
    nb.create_tests(
        tag="private", oktest_path=base_name, base_dir=autograder_path
    )
    nb.create_tests(tag="public", oktest_path=base_name, base_dir=student_path)
    text_replace_begin = "### BEGIN SOLUTION"
    text_replace_end = "### END SOLUTION"
    nb.replace_text(text_replace_begin, text_replace_end)

    nb.save(os.path.join(student_path, nb_name))
    normalize_kernel_name(os.path.join(student_path, nb_name))

    # create test files for the autograder
    nb = NotebookCleaner(notebook)
    nb.create_tests(
        tag="private", oktest_path=base_name, base_dir=autograder_path
    )
    nb.create_tests(
        tag="public", oktest_path=base_name, base_dir=autograder_path
    )


def find_check_definition(tree):
    """Walk an AST and check for definitions of a function called `check`

    Return True if one is found, False otherwise.
    """
    for stmt in ast.walk(tree):
        if not isinstance(stmt, ast.FunctionDef):
            continue
        if stmt.name == "check":
            return True
    return False


def find_check_assignment(tree):
    """Walk an AST and check for assignments to a variable called `check`

    Return True if one is found, False otherwise.
    """
    for stmt in ast.walk(tree):
        if not isinstance(stmt, ast.Assign):
            continue
        # check id for tuple target
        target_names = []
        for target in stmt.targets:
            if isinstance(target, tuple):
                target_names += [t.id for t in target]
            else:
                target_names.append(target.id)
        if "check" in target_names:
            return True
    return False


def execute_notebook(nb_path):
    """Execute a notebook under grading conditions"""
    graded_nb_path = os.path.splitext(nb_path)[0] + "-graded.ipynb"
    nb_directory = os.path.split(nb_path)[0]

    # read in input notebook and check the source for shenanigans
    nb = nbformat.read(nb_path, as_version=4)
    source = ""
    for cell in nb.cells:
        if cell.cell_type != "code":
            continue

        isp = IPythonInputSplitter(line_input_checker=False)
        cell_source = isp.transform_cell(cell.source)
        source += cell_source

    tree = ast.parse(source)

    # no points for you if you try and cheat
    # XXX add a check for people importing a function called `check`
    if find_check_assignment(tree) or find_check_definition(tree):
        return

    # run the notebook
    with chdir(nb_directory):
        pm.execute_notebook(nb_path, graded_nb_path)

    graded_nb = nbformat.read(graded_nb_path, as_version=4)
    return graded_nb



"""
abc-classroom.feedback
======================

"""

from pathlib import Path
import csv
import shutil

from . import config as cf
from . import github


def copy_feedback(args):
    """
    Copies feedback reports to local student repositories, commits the changes, and (optionally) pushes to github. Assumes files are in the directory course_materials/feedback/student/assignment. Copies all files in the source directory.
    """
    assignment = args.assignment
    do_github = args.github

    print("Loading configuration from config.yml")
    config = cf.get_config()

    # get various paths from config
    roster_filename = cf.get_config_option(config, "roster", True)
    course_dir = cf.get_config_option(config, "course_directory", True)
    clone_dir = cf.get_config_option(config, "clone_dir", True)
    organization = cf.get_config_option(config, "organization", True)
    materials_dir = cf.get_config_option(config, "course_materials", True)

    try:
        feedback_dir = Path(course_dir, materials_dir, "feedback")
        with open(roster_filename, newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                student = row["github_username"]
                source_files = Path(feedback_dir, student, assignment).glob(
                    "*"
                )
                repo = "{}-{}".format(assignment, student)
                destination_dir = Path(clone_dir, repo)
                if not destination_dir.is_dir():
                    print(
                        "Local student repository {} does not exist; skipping student".format(
                            destination_dir
                        )
                    )
                    continue
                for f in source_files:
                    print(
                        "Copying {} to {}".format(
                            f.relative_to(course_dir), destination_dir
                        )
                    )
                    shutil.copy(f, destination_dir)
                github.commit_all_changes(
                    destination_dir,
                    msg="Adding feedback for assignment {}".format(assignment),
                )
                if do_github:
                    github.push_to_github(destination_dir)

    except FileNotFoundError as err:
        print("Missing file or directory:")
        print(" ", err)



"""
abc-classroom.distribute
========================

Take a template directory and distribute it
"""
import os
import subprocess

import jinja2


# current directory
HERE = os.path.dirname(os.path.abspath(__file__))


def find_notebooks(base_path):
    """Find all notebooks below base_path and return path to them"""
    notebook_paths = []
    for path, _, files in os.walk(base_path):
        # remove leading directory
        path = os.path.relpath(path, base_path)
        for f in files:
            if f.endswith(".ipynb"):
                notebook_paths.append(os.path.join(path, f))

    return notebook_paths


def render_circleci_template(notebook_paths):
    """Load the CircleCI template"""
    template = open(os.path.join(HERE, "circleci.yml")).read()

    t = jinja2.Template(template)

    return t.render(
        notebook_dirs=[os.path.dirname(p) for p in notebook_paths],
        notebook_names=[os.path.split(p)[1] for p in notebook_paths],
        notebooks=notebook_paths,
    )


def git_init(directory):
    """Initialise and populate a git repository in `directory`

    Add all files in `directory` to a newly create git repository in that
    directory.
    """
    subprocess.run(
        ["git", "init"], cwd=directory, check=True, stdout=subprocess.PIPE
    )

    subprocess.run(
        ["git", "add", "*"], cwd=directory, check=True, stdout=subprocess.PIPE
    )

    subprocess.run(
        ["git", "commit", "-m", "Initial commit"],
        cwd=directory,
        check=True,
        stdout=subprocess.PIPE,
    )



"""
abc-classroom.ok
================

"""


import doctest
import inspect
import io
import os
from contextlib import redirect_stderr, redirect_stdout
from jinja2 import Template
from textwrap import dedent

from .notebook import execute_notebook
from .utils import hide_outputs
from pygments import highlight
from pygments.lexers import PythonConsoleLexer
from pygments.formatters import HtmlFormatter


def run_doctest(name, doctest_string, global_environment):
    """
    Run a single test with given global_environment.
    Returns (True, '') if the doctest passes.
    Returns (False, failure_message) if the doctest fails.
    """
    examples = doctest.DocTestParser().parse(doctest_string, name)
    test = doctest.DocTest(
        [e for e in examples if isinstance(e, doctest.Example)],
        global_environment,
        name,
        None,
        None,
        doctest_string,
    )

    doctestrunner = doctest.DocTestRunner(verbose=True)

    runresults = io.StringIO()
    with redirect_stdout(runresults), redirect_stderr(
        runresults
    ), hide_outputs():
        doctestrunner.run(test, clear_globs=False)
    with open("/dev/null", "w") as f, redirect_stderr(f), redirect_stdout(f):
        result = doctestrunner.summarize(verbose=True)
    # An individual test can only pass or fail
    if result.failed == 0:
        return (True, "")
    else:
        return False, runresults.getvalue()


class OKTest:
    """
    A single DocTest defined by OKPy.
    Instances of this class are callable. When called, it takes
    a global_environment dict, and returns a TestResult object.
    We only take a global_environment, *not* a local_environment.
    This makes tests not useful inside functions, methods or
    other scopes with local variables. This is a limitation of
    doctest, so we roll with it.
    """

    result_pass_template = Template(
        """
    <p><strong>{{ name }}</strong> passed!</p>
    """
    )

    result_fail_template = Template(
        """
    <p><strong style='color: red;'>{{ name }}</strong></p>
    <p><strong>Test code:</strong><pre>{{test_code}}</pre></p>
    <p><strong>Test result:</strong><pre>{{test_result}}</pre></p>
    """
    )

    def __init__(self, name, tests):
        """
        tests is list of doctests that should be run.
        """
        self.name = name
        self.tests = tests

    def run(self, global_environment):
        results = []
        for i, (max_points, test) in enumerate(self.tests):
            name = "{} test #{}".format(self.name, i + 1)
            passed, result = run_doctest(name, test, global_environment)
            if not passed:
                results.append(
                    (
                        False,
                        (0, max_points),
                        OKTest.result_fail_template.render(
                            name=name,
                            test_code=highlight(
                                test,
                                PythonConsoleLexer(),
                                HtmlFormatter(noclasses=True),
                            ),
                            test_result=result,
                        ),
                    )
                )
            else:
                results.append(
                    (
                        True,
                        (max_points, max_points),
                        OKTest.result_pass_template.render(name=name),
                    )
                )

        return results

    @classmethod
    def from_file(cls, path):
        """
            Parse a ok test file & return an OKTest
        """
        # ok test files are python files, with a global 'test' defined
        test_globals = {}
        with open(path) as f:
            exec(f.read(), test_globals)

        test_spec = test_globals["test"]

        # We only support a subset of these tests, so let's validate!

        # Make sure there is a name
        assert "name" in test_spec

        # Do not support multiple suites in the same file
        assert len(test_spec["suites"]) == 1

        # Do not support point values other than 1
        assert test_spec.get("points", 1) == 1

        test_suite = test_spec["suites"][0]

        # Only support doctest. I am unsure if other tests are implemented
        assert test_suite.get("type", "doctest") == "doctest"

        # Not setup and teardown supported
        assert not bool(test_suite.get("setup"))
        assert not bool(test_suite.get("teardown"))

        tests = []

        for i, test_case in enumerate(test_spec["suites"][0]["cases"]):
            tests.append((int(test_case["points"]), dedent(test_case["code"])))

        return cls(path, tests)


class OKSuite:
    def __init__(self, test_paths):
        self.tests = [OKTest.from_file(path) for path in test_paths]

    def run(self, global_environment, include_grade=True):
        passed_tests = []
        failed_tests = []
        points_scored = 0
        max_total_points = 0
        for t in self.tests:
            for passed, points, hint in t.run(global_environment):
                points_scored += points[0]
                max_total_points += points[1]
                if passed:
                    passed_tests.append(hint)
                else:
                    failed_tests.append(hint)

        grade = len(passed_tests) / len(passed_tests + failed_tests)
        return OKSuiteResult(
            (points_scored, max_total_points),
            passed_tests,
            failed_tests,
            include_grade,
        )


class OKSuiteResult:
    """
    Displayable result from running OKTests
    """

    result_template = Template(
        """
    {% if include_grade %}
    <strong>Grade: {{ grade }}</strong>
    {% endif %}
    {% if grade[0] == grade[1] %}
        <p>All {{ tests|length }} tests passed! Points: {{ grade[0] }}.</p>
    {% else %}
        <p>{{ passed_tests|length }} of {{ tests|length }} tests passed. Points: {{ grade[0] }} of {{ grade[1] }}.</p>
        {% if passed_tests %}
        <p> <strong>Tests passed:</strong>
            <ul>
            {% for passed_test_hint in passed_tests %}
                <li> {{ passed_test_hint }} </li>
            {% endfor %}
            </ul>
        </p>
        {% endif %}
        {% if failed_tests %}
        <p> <strong>Tests failed: </strong>
            <ul>
            {% for failed_test_hint in failed_tests %}
                <li> {{ failed_test_hint }} </li>
            {% endfor %}
            </ul>
        {% endif %}
    {% endif %}
    """
    )

    def __init__(self, grade, passed_tests, failed_tests, include_grade=True):
        self.grade = grade
        self.passed_tests = passed_tests
        self.failed_tests = failed_tests
        self.tests = passed_tests + failed_tests
        self.include_grade = include_grade

    def _repr_html_(self):
        return OKSuiteResult.result_template.render(
            grade=self.grade,
            passed_tests=self.passed_tests,
            failed_tests=self.failed_tests,
            tests=self.tests,
            include_grade=self.include_grade,
        )


def grade_notebook(notebook_path):
    executed_nb = execute_notebook(notebook_path)
    # collect marks from each cell containing a `check()` call
    results = []

    for cell in executed_nb.cells:
        if cell.cell_type != "code":
            continue

        if cell.source.startswith("check(") and cell.source.endswith(")"):
            output = cell["outputs"][0]["data"]["text/html"].strip()
            # got all points
            if "tests passed! Points:" in output:
                print(output)
                points = output.rsplit(" ")[-1]
                points = int(points[: -len(".</p>")])
                results.append((points, points))
            # got some of the points
            elif "tests passed. Points:" in output and "of" in output:
                output = output.split("\n")[0]
                print(output)
                out = output.rsplit(" ")
                scored = int(out[-3])
                total = int(out[-1][: -len(".</p>")])
                results.append((scored, total))

            if output.endswith("Full grade.</p>"):
                results.append(1.0)

    return results


def check(test_file_path, global_env=None):
    """
    check global_env against given test_file in oktest format
    If global_env is none, the global environment of the calling
    function is used. The following two calls are equivalent:
    check('tests/q1.py')
    check('tests/q1.py', globals())
    Returns a TestResult object.
    """
    # It isn't an error to call this with a non-existant file
    # students don't have access to all tests
    if not os.path.exists(test_file_path):
        return None

    tests = OKSuite([test_file_path])

    if global_env is None:
        # Get the global env of our callers - one level below us in the stack
        # The grade method should only be called directly from user / notebook
        # code. If some other method is calling it, it should also use the
        # inspect trick to pass in its parents' global env.
        global_env = inspect.currentframe().f_back.f_globals
    return tests.run(global_env, include_grade=False)



import os

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

DISTNAME = "abc-classroom"
DESCRIPTION = (
    "Efficiently manage github classroom assignments from the command line."
)
MAINTAINER = "Leah Wasser"
MAINTAINER_EMAIL = "leah.wasser@colorado.edu"

setup(
    name=DISTNAME,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version="0.0.15",
    license="BSD",
    packages=["abcclassroom"],
    install_requires=[
        "nbclean",
        "jinja2",
        "papermill",
        "nbformat",
        "ruamel.yaml",
        "github3.py",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "abc-init = abcclassroom.__main__:init",
            "abc-quickstart = abcclassroom.__main__:quickstart",
            "abc-new-template = abcclassroom.__main__:new_template",
            "abc-update-template = abcclassroom.__main__:update_template",
            "abc-clone = abcclassroom.__main__:clone",
            "abc-feedback = abcclassroom.__main__:feedback",
        ]
    },
    url="https://github.com/earthlab/abc-classroom",
)



"""
abc-classroom.template
======================

"""
import os
import sys
import shutil
import copy
from pathlib import Path

from . import config as cf
from . import github
from . import utils


def new_update_template(args):
    """
    Creates or updates an assignment template repository. Implementation of both the new_template and update_template console scripts (which perform the same basic functions but with different command line arguments and defaults).

    Creates an assignment entry in the config file if one does not already exist.
    """
    print("Loading configuration from config.yml")
    config = cf.get_config()
    template_dir = cf.get_config_option(config, "template_dir", True)

    # create the local git repository
    assignment = args.assignment
    template_repo_path = create_template_dir(config, assignment, args.mode)
    print("repo path: {}".format(template_repo_path))
    copy_assignment_files(config, template_repo_path, assignment)
    create_extra_files(config, template_repo_path, assignment)
    github.init_and_commit(template_repo_path, args.custom_message)

    # create / append assignment entry in config
    course_dir = cf.get_config_option(config, "course_directory", True)
    cf.print_config(config)
    cf.set_config_option(
        config,
        "assignments",
        assignment,
        append_value=True,
        configpath=course_dir,
    )

    # optional github steps
    if args.github:
        organization = cf.get_config_option(config, "organization", True)
        repo_name = os.path.basename(template_repo_path)
        token = cf.get_github_auth()["token"]

        create_or_update_remote(
            template_repo_path, organization, repo_name, token
        )


def create_or_update_remote(
    template_repo_path, organization, repo_name, token
):
    remote_exists = github.remote_repo_exists(organization, repo_name, token)
    if not remote_exists:
        print("Creating remote repo {}".format(repo_name))
        # create the remote repo on github and push the local repo
        # (will print error and return if repo already exists)
        github.create_repo(organization, repo_name, token)

    try:
        github.add_remote(template_repo_path, organization, repo_name, token)
    except RuntimeError as e:
        print("Remote already added to local repository")
        pass

    print("Pushing any changes to remote repository on GitHub")
    try:
        github.push_to_github(template_repo_path, "master")
    except RuntimeError as e:
        print(
            "Push to github failed. This is usually because there are changes on the remote that you do not have locally. Here is the github error:"
        )
        print(e)


def create_template_dir(config, assignment, mode="fail"):
    """
    Creates a new directory in template_dir that will become the
    template repository for the assignment. If directory exists and mode is  merge, do nothing. If directory exists and mode is delete, remove contents but leave .git directory.
    """
    course_dir = cf.get_config_option(config, "course_directory", True)
    template_parent_dir = cf.get_config_option(config, "template_dir", True)
    parent_path = utils.get_abspath(template_parent_dir, course_dir)

    # check that parent directory for templates exists, and create it if it does not
    if not os.path.isdir(parent_path):
        print(
            "Creating new directory for template repos at {}".format(
                parent_path
            )
        )
        os.mkdir(parent_path)

    repo_name = assignment + "-template"
    template_path = Path(parent_path, repo_name)
    dir_exists = template_path.is_dir()
    if not dir_exists:
        template_path.mkdir()
        print("Creating new template repo at {}".format(template_path))
    else:
        if mode == "fail":
            print(
                "Directory {} already exists; re-run with '--mode merge' or '--mode delete', or delete / move directory before re-running".format(
                    template_path
                )
            )
            sys.exit(1)
        elif mode == "merge":
            print(
                "Template directory {} already exists; will keep directory but overwrite existing files with same names".format(
                    template_path
                )
            )
        else:
            # mode == delete
            print(
                "Template directory {} already exists; deleting existing files but keeping .git directory, if exists.".format(
                    template_path
                )
            )
            # temporarily move the .git dir to the parent of the template_path
            gitdir = Path(template_path, ".git")
            if gitdir.exists():
                target = Path(Path(template_path).parent, ".tempgit")
                gitdir.replace(target)

                # remove template_path and re-create with same name
                shutil.rmtree(template_path)
                Path(template_path).mkdir()

                # and then move the .git dir back
                target.replace(gitdir)

    return template_path


def copy_assignment_files(config, template_repo, assignment):
    """Copy all of the files from the course_materials/release directory for the
    assignment into the template repo directory.
    """

    print("Getting assignment files")
    course_dir = cf.get_config_option(config, "course_directory", True)
    materials_dir = cf.get_config_option(config, "course_materials", True)
    parent_path = utils.get_abspath(materials_dir, course_dir)
    release_dir = os.path.join(parent_path, "release", assignment)
    if not os.path.exists(release_dir):
        print(
            "release directory {} does not exist; exiting\n".format(
                release_dir
            )
        )
        sys.exit(1)
    nfiles = 0
    all_files = os.listdir(release_dir)
    # matched_files = match_patterns(all_files, patterns)
    for file in all_files:
        fpath = os.path.join(release_dir, file)
        print("copying {} to {}".format(fpath, template_repo))
        # overwrites if fpath exists in template_repo
        shutil.copy(fpath, template_repo)
        nfiles += 1
    print("Copied {} files".format(nfiles))


def create_extra_files(config, template_repo, assignment):
    """Create any extra files as specified in the config """
    extra_files = copy.deepcopy(
        cf.get_config_option(config, "extra_files", False)
    )
    nfiles = len(extra_files)
    print("Creating {} extra files".format(nfiles))
    for file, contents in extra_files.items():
        if len(contents) > 0:
            if file == "README.md":
                firstline = ""
                if assignment:
                    first_line = "# {}\n".format(assignment)
                else:
                    first_line = "# README\n"
                contents.insert(0, first_line)
            utils.write_file(template_repo, file, contents)
